@@ -6,20 +6,19 @@ Created on Wed Sep  11 12:04:23 2021
 @author: daniel
 """
 import sys 
-from warnings import filterwarnings
-filterwarnings("ignore", category=FutureWarning)
-import numpy as np
 import random
+import numpy as np
+from pandas import DataFrame
+from warnings import filterwarnings, warn
+filterwarnings("ignore", category=FutureWarning)
 
 from sklearn.impute import KNNImputer
-from sklearn.metrics import accuracy_score
 from sklearn.ensemble import RandomForestClassifier
 from sklearn.neural_network import MLPClassifier
 from sklearn.model_selection import train_test_split, cross_validate, cross_val_score
 import sklearn.neighbors._base
 sys.modules['sklearn.neighbors.base'] = sklearn.neighbors._base
 from missingpy import MissForest
-from boruta import BorutaPy
 
 from skopt import BayesSearchCV, plots, gp_minimize
 from skopt.plots import plot_convergence, plot_objective
@@ -27,45 +26,51 @@ from skopt.space import Real, Integer, Categorical
 from skopt.utils import use_named_args
 
 import optuna
-optuna.logging.set_verbosity(optuna.logging.WARN)
+from boruta import BorutaPy
+from BorutaShap import BorutaShap
+optuna.logging.set_verbosity(optuna.logging.WARNING)
 from xgboost import XGBClassifier
 
-def objective_xgb(trial, data_x, data_y):
+class objective_xgb(object):
     """
     The Optuna objective function for the tree-based XGBoost classifier. 
     The Optuna software for hyperparameter optimization was published in 
     2019 by Akiba et al. Paper: https://arxiv.org/abs/1907.10902
     """
+    def __init__(self, data_x, data_y):
+        self.data_x = data_x
+        self.data_y = data_y
 
-    booster = trial.suggest_categorical('booster', ['gbtree', 'dart'])
-    reg_lambda = trial.suggest_loguniform('reg_lambda', 1e-6, 1)
-    reg_alpha = trial.suggest_loguniform('reg_alpha', 1e-6, 1)
-    max_depth = trial.suggest_int('max_depth', 1, 9)
-    eta = trial.suggest_loguniform('eta', 1e-6, 1)
-    gamma = trial.suggest_loguniform('gamma', 1e-6, 1)
-    grow_policy = trial.suggest_categorical('grow_policy', ['depthwise', 'lossguide'])
+    def __call__(self, trial):
+        booster = trial.suggest_categorical('booster', ['gbtree', 'dart'])
+        reg_lambda = trial.suggest_loguniform('reg_lambda', 1e-6, 1)
+        reg_alpha = trial.suggest_loguniform('reg_alpha', 1e-6, 1)
+        max_depth = trial.suggest_int('max_depth', 1, 9)
+        eta = trial.suggest_loguniform('eta', 1e-6, 1)
+        gamma = trial.suggest_loguniform('gamma', 1e-6, 1)
+        grow_policy = trial.suggest_categorical('grow_policy', ['depthwise', 'lossguide'])
 
-    if booster == "dart":
-        sample_type = trial.suggest_categorical('sample_type', ['uniform', 'weighted'])
-        normalize_type = trial.suggest_categorical('normalize_type', ['tree', 'forest'])
-        rate_drop = trial.suggest_loguniform('rate_drop', 1e-6, 1.0)
-        skip_drop = trial.suggest_loguniform('skip_drop', 1e-6, 1.0)
+        if booster == "dart":
+            sample_type = trial.suggest_categorical('sample_type', ['uniform', 'weighted'])
+            normalize_type = trial.suggest_categorical('normalize_type', ['tree', 'forest'])
+            rate_drop = trial.suggest_loguniform('rate_drop', 1e-6, 1.0)
+            skip_drop = trial.suggest_loguniform('skip_drop', 1e-6, 1.0)
 
-        clf = XGBClassifier(booster=booster, reg_lambda=reg_lambda, reg_alpha=reg_alpha, max_depth=max_depth, eta=eta, 
-            gamma=gamma, grow_policy=grow_policy, sample_type=sample_type, normalize_type=normalize_type,rate_drop=rate_drop, 
-            skip_drop=skip_drop)
-    
-    elif booster == 'gbtree':
-        subsample = trial.suggest_loguniform('subsample', 1e-6, 1.0)
-        clf = XGBClassifier(booster=booster, reg_lambda=reg_lambda, reg_alpha=reg_alpha, max_depth=max_depth, eta=eta, 
-            gamma=gamma, grow_policy=grow_policy, subsample=subsample)
+            clf = XGBClassifier(booster=booster, reg_lambda=reg_lambda, reg_alpha=reg_alpha, max_depth=max_depth, eta=eta, 
+                gamma=gamma, grow_policy=grow_policy, sample_type=sample_type, normalize_type=normalize_type,rate_drop=rate_drop, 
+                skip_drop=skip_drop)
+        
+        elif booster == 'gbtree':
+            subsample = trial.suggest_loguniform('subsample', 1e-6, 1.0)
+            clf = XGBClassifier(booster=booster, reg_lambda=reg_lambda, reg_alpha=reg_alpha, max_depth=max_depth, eta=eta, 
+                gamma=gamma, grow_policy=grow_policy, subsample=subsample)
 
-    cv = cross_validate(clf, data_x, data_y, cv=3)
-    final_score = np.round(np.mean(cv['test_score']), 4)
+        cv = cross_validate(clf, self.data_x, self.data_y, cv=3)
+        final_score = np.round(np.mean(cv['test_score']), 6)
 
-    return final_score
+        return final_score
 
-def objective_nn(trial, data_x, data_y):
+class objective_nn(object):
     """
     The Optuna objective function for the scikit-learn implementatin of the
     MLP classifier. The Optuna software for hyperparameter optimization
@@ -74,49 +79,59 @@ def objective_nn(trial, data_x, data_y):
     See Optuna examples: https://github.com/optuna/optuna-examples 
     """
 
-    learning_rate_init= trial.suggest_float('learning_rate_init', 1e-6, 1)
-    solver = trial.suggest_categorical("solver", ["sgd", "adam"]) #"lbfgs"
-    activation = trial.suggest_categorical("activation", ["identity", "logistic", "tanh", "relu"])
-    learning_rate = trial.suggest_categorical("learning_rate", ["constant", "invscaling", "adaptive"])
-    alpha = trial.suggest_float("alpha", 1e-6, 1)
-    batch_size = trial.suggest_int('batch_size', 1, 1000)
-    
-    n_layers = trial.suggest_int('hidden_layer_sizes', 1, 10)
-    layers = []
-    for i in range(n_layers):
-        layers.append(trial.suggest_int(f'n_units_{i}', 100, 1000))
+    def __init__(self, data_x, data_y):
+        self.data_x = data_x
+        self.data_y = data_y
 
-    clf = MLPClassifier(hidden_layer_sizes=tuple(layers),learning_rate_init=learning_rate_init, 
-        solver=solver, activation=activation, alpha=alpha, batch_size=batch_size, max_iter=2500)
-    
-    cv = cross_validate(clf, data_x, data_y, cv=3)
-    final_score = np.round(np.mean(cv['test_score']), 4)
+    def __call__(self, trial):
+        learning_rate_init= trial.suggest_float('learning_rate_init', 1e-6, 1)
+        solver = trial.suggest_categorical("solver", ["sgd", "adam"]) #"lbfgs"
+        activation = trial.suggest_categorical("activation", ["logistic", "tanh", "relu"])
+        learning_rate = trial.suggest_categorical("learning_rate", ["constant", "invscaling", "adaptive"])
+        alpha = trial.suggest_float("alpha", 1e-6, 1)
+        batch_size = trial.suggest_int('batch_size', 1, 1000)
+        
+        n_layers = trial.suggest_int('hidden_layer_sizes', 1, 10)
+        layers = []
+        for i in range(n_layers):
+            layers.append(trial.suggest_int(f'n_units_{i}', 100, 1000))
 
-    return final_score
+        clf = MLPClassifier(hidden_layer_sizes=tuple(layers),learning_rate_init=learning_rate_init, 
+            solver=solver, activation=activation, alpha=alpha, batch_size=batch_size, max_iter=2500)
+        
+        cv = cross_validate(clf, self.data_x, self.data_y, cv=3)
+        final_score = np.round(np.mean(cv['test_score']), 6)
 
-def objective_rf(trial, data_x, data_y):
+        return final_score
+
+class objective_rf(object):
     """
     The Optuna objective function for the scikit-learn implementatin of the
     Random Forest classifier. The Optuna software for hyperparameter optimization
     was published in 2019 by Akiba et al. Paper: https://arxiv.org/abs/1907.10902
     """
 
-    n_estimators = trial.suggest_int('n_estimators', 100, 3000)
-    criterion = trial.suggest_categorical('criterion', ['gini', 'entropy'])
-    max_depth = trial.suggest_int('max_depth', 1, 100)
-    min_samples_split = trial.suggest_float('min_samples_split', 0,1)
-    min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 10)
-    max_features = trial.suggest_int('max_features', 1, data_x.shape[1])
-    bootstrap = trial.suggest_categorical('bootstrap', [True, False])
-    
-    clf = RandomForestClassifier(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth,
-                                min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
-                                max_features=max_features, bootstrap=bootstrap)
-    
-    cv = cross_validate(clf, data_x, data_y, cv=3)
-    final_score = np.round(np.mean(cv['test_score']), 4)
+    def __init__(self, data_x, data_y):
+        self.data_x = data_x
+        self.data_y = data_y
 
-    return final_score
+    def __call__(self, trial):
+        n_estimators = trial.suggest_int('n_estimators', 100, 3000)
+        criterion = trial.suggest_categorical('criterion', ['gini', 'entropy'])
+        max_depth = trial.suggest_int('max_depth', 1, 25)
+        min_samples_split = trial.suggest_int('min_samples_split', 1, 25)
+        min_samples_leaf = trial.suggest_int('min_samples_leaf', 1, 15)
+        max_features = trial.suggest_int('max_features', 1, self.data_x.shape[1])
+        bootstrap = trial.suggest_categorical('bootstrap', [True, False])
+        
+        clf = RandomForestClassifier(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth,
+                                    min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
+                                    max_features=max_features, bootstrap=bootstrap)
+        
+        cv = cross_validate(clf, self.data_x, self.data_y, cv=3)
+        final_score = np.round(np.mean(cv['test_score']), 6)
+
+        return final_score
 
 def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
     """
@@ -174,6 +189,7 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
     elif clf == 'nn':
         model_0 = MLPClassifier()
     elif clf == 'xgb':
+        model_0 = XGBClassifier()
         if all(isinstance(val, (int, str)) for val in data_y):
             print('XGBoost classifier requires numerical class labels! Converting class labels as follows:')
             print('____________________________________')
@@ -183,27 +199,21 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
                 index = np.where(data_y == np.unique(data_y)[i])[0]
                 y[index] = i
             data_y = y 
-            print('____________________________________')
-        model_0 = XGBClassifier()
+            print('------------------------------------')
     else:
         raise ValueError('clf argument must either be "rf", "nn", or "xgb".')
 
     cv = cross_validate(model_0, data_x, data_y, cv=3)
-    initial_score = np.round(np.mean(cv['test_score']), 4)
-
-    def logging_callback(study, trial):
-        previous_best = study.user_attrs.get("previous_best", None)
-        if previous_best != study.best_value:
-            study.set_user_attr("previous_best", study.best_value)
-            print("Highest 3-fold CV accuracy: {} (Trial # {}). ".format(trial.value, trial.number))
+    initial_score = np.round(np.mean(cv['test_score']), 6)
 
     sampler = optuna.samplers.TPESampler(seed=1909) 
     study = optuna.create_study(direction='maximize', sampler=sampler)
-
-    print('Running hyperparameter optimization procedure, this will take a while...')
+    print('Starting hyperparameter optimization, this will take a while...')
     if clf == 'rf':
         try:
-            study.optimize(lambda trial: objective_rf(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
+            objective = objective_rf(data_x, data_y)
+            study.optimize(objective, n_trials=n_iter, show_progress_bar=True)
+            #study.optimize(lambda trial: objective_rf(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
             params = study.best_trial.params
             model = RandomForestClassifier(n_estimators=params['n_estimators'], criterion=params['criterion'], 
                 max_depth=params['max_depth'], min_samples_split=params['min_samples_split'], 
@@ -230,7 +240,9 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
 
     elif clf == 'nn':
         try:
-            study.optimize(lambda trial: objective_nn(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
+            objective = objective_nn(data_x, data_y)
+            study.optimize(objective, n_trials=n_iter, show_progress_bar=True)
+            #study.optimize(lambda trial: objective_nn(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
             params = study.best_trial.params
             layers = [param for param in params if 'n_units_' in param]
             layers = tuple(params[layer] for layer in layers)
@@ -238,26 +250,38 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
                 activation=params['activation'], learning_rate=params['learning_rate'], alpha=params['alpha'], 
                 batch_size=params['batch_size'], solver=params['solver'], max_iter=2500)
         except:
-            raise ValueError('Error occurred while optimizing neural network with Optuna.')
-
+            print('Failed to optimize with Optuna, switching over to BayesSearchCV...')
+            params = {
+                'hidden_layer_sizes': [(100,),(50,100,50),(75,50,20),(150,100,50),(120,80,40),(100,50,30)],
+                'learning_rate': ['constant', 'invscaling', 'adaptive'],
+                'alpha': [0.00001, 0.5], 
+                'activation': ['tanh', 'logistic', 'relu'],
+                'solver': ['sgd', 'adam'],
+                'max_iter': [100, 150, 200] 
+            }
+            gs = BayesSearchCV(n_iter=n_iter, estimator=MLPClassifier(), search_spaces=params, cv=3)
+            gs.fit(data_x, data_y)
+            best_est, best_score = gs.best_estimator_, np.round(gs.best_score_, 4)
+            print('Highest mean accuracy: {}'.format(best_score))
+            return gs.best_estimator_, gs.best_params_
+          
     elif clf == 'xgb':
-        try:
-            study.optimize(lambda trial: objective_xgb(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
-            params = study.best_trial.params
-            if params['booster'] == 'dart':
-                model = XGBClassifier(booster=params['booster'], reg_lambda=params['reg_lambda'], reg_alpha=params['reg_alpha'], 
-                    max_depth=params['max_depth'], eta=params['eta'], gamma=params['gamma'], grow_policy=params['grow_policy'],
-                    sample_type=params['sample_type'], normalize_type=params['normalize_type'],rate_drop=params['rate_drop'], 
-                    skip_drop=params['skip_drop'])
-            elif params['booster'] == 'gbtree':
-                model = XGBClassifier(booster=params['booster'], reg_lambda=params['reg_lambda'], reg_alpha=params['reg_alpha'], 
-                    max_depth=params['max_depth'], eta=params['eta'], gamma=params['gamma'], grow_policy=params['grow_policy'],
-                    subsample=params['subsample'], sampling_method=params['sampling_method'], tree_method=params['tree_method'],
-                    process_type=params['process_type'])
-        except:
-            raise ValueError('Error occurred while optimizing XGBoost with Optuna.')
-
-    final_score = study.best_trial.value
+        objective = objective_xgb(data_x, data_y)
+        study.optimize(objective, n_trials=n_iter, show_progress_bar=True)
+        #study.optimize(lambda trial: objective_xgb(trial, data_x=data_x, data_y=data_y), n_trials=n_iter, callbacks=[logging_callback])
+        params = study.best_trial.params
+        if params['booster'] == 'dart':
+            model = XGBClassifier(booster=params['booster'], reg_lambda=params['reg_lambda'], reg_alpha=params['reg_alpha'], 
+                max_depth=params['max_depth'], eta=params['eta'], gamma=params['gamma'], grow_policy=params['grow_policy'],
+                sample_type=params['sample_type'], normalize_type=params['normalize_type'],rate_drop=params['rate_drop'], 
+                skip_drop=params['skip_drop'])
+        elif params['booster'] == 'gbtree':
+            model = XGBClassifier(booster=params['booster'], reg_lambda=params['reg_lambda'], reg_alpha=params['reg_alpha'], 
+                max_depth=params['max_depth'], eta=params['eta'], gamma=params['gamma'], grow_policy=params['grow_policy'],
+                subsample=params['subsample'], sampling_method=params['sampling_method'], tree_method=params['tree_method'],
+                process_type=params['process_type'])
+       
+    final_score = study.best_value
 
     if initial_score > final_score:
         print('Hyperparameter optimization complete! 3-fold CV accuracy of {} is lower than the base accuracy of {}, try increasing the value of n_iter and run again.'.format(np.round(final_score, 4), np.round(initial_score, 4)))
@@ -268,6 +292,66 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, save_study=False):
         return model, params, study
     return model, params
 
+def borutashap_opt(data_x, data_y, model='rf', n_trials=50):
+    """
+    Applies a combination of the Boruta algorithm and
+    Shapley values, a method developed by Eoghan Keany (2020).
+
+    See: https://doi.org/10.5281/zenodo.4247618
+
+    Args:
+        data_x (ndarray): 2D array of size (n x m), where n is the
+            number of samples, and m the number of features.
+        data_y (ndarray, str): 1D array containing the corresponing labels.
+        model (str): The ensemble method to use when fitting and calculating
+            the feature importance metric. Only two options are currently
+            supported, 'rf' for Random Forest and 'xgb' for Extreme Gradient Boosting.
+            Defaults to 'rf'.
+        n_trials (int): The number of trials to run. A larger number is
+            better as the distribution will be more robust to random fluctuations. 
+            Defaults to 50.
+    Returns:
+        1D array containing the indices of the selected features. This can then
+        be used to index the columns in the data_x array.
+    """
+    if n_trials < 20:
+        warn('Results are unstable if n_trials is too low!')
+    if np.any(np.isnan(data_x)):
+        print('NaN values detected, applying Strawman imputation...')
+        data_x = Strawman_imputation(data_x)
+
+    if model == 'rf':
+        classifier = RandomForestClassifier()
+    elif model == 'xgb':
+        classifier = XGBClassifier()
+    else:
+        raise ValueError('Model argument must either be "rf" or "xgb".')
+    
+    try:
+        #BorutaShap program requires input to have the columns attribute
+        #Converting to Pandas dataframe
+        cols = [str(i) for i in np.arange(data_x.shape[1])]
+        X = DataFrame(data_x, columns=cols)
+        y = np.zeros(len(data_y))
+
+        #Below is to convert categorical labels to numerical, as per BorutaShap requirements
+        for i, label in enumerate(np.unique(data_y)):
+            mask = np.where(data_y == label)[0]
+            y[mask] = i
+
+        feat_selector = BorutaShap(model=classifier, importance_measure='shap', classification=True)
+        print('Running feature selection...')
+        feat_selector.fit(X=X, y=y, n_trials=n_trials, verbose=False, random_state=1909)
+
+        index = np.array([int(feat) for feat in feat_selector.accepted])
+        index.sort()
+        print('Feature selection complete, {} selected out of {}!'.format(len(index), data_x.shape[1]))
+    except:
+        print('Boruta with Shapley values failed, switching to original Boruta...')
+        index = boruta_opt(data_x, data_y)
+
+    return index
+
 def boruta_opt(data_x, data_y):
     """
     Applies the Boruta algorithm (Kursa & Rudnicki 2011) to identify features
@@ -287,21 +371,14 @@ def boruta_opt(data_x, data_y):
 
     classifier = RandomForestClassifier()
 
-    feat_selector = BorutaPy(classifier, n_estimators='auto', verbose=0, random_state=1)
-    print('Running feature optimization...')
+    feat_selector = BorutaPy(classifier, n_estimators='auto', random_state=1909)
+    print('Running feature selection...')
     feat_selector.fit(data_x, data_y)
 
-    feat_selector.support = np.array([str(feat) for feat in feat_selector.support_])
-    index = np.where(feat_selector.support == 'True')[0]
+    feats = np.array([str(feat) for feat in feat_selector.support_])
+    index = np.where(feats == 'True')[0]
     print('Feature selection complete, {} selected out of {}!'.format(len(index),len(feat_selector.support)))
     return index
-    """
-    cv = cross_validate(classifier, data_x[:,index], data_y, cv=10)
-    print('Accuracy without feature optimization: {}'.format(np.round(np.mean(cv['test_score']),3)))
-    print('-------------------------------------')
-    cv = cross_validate(classifier, data_x[:,index], data_y, cv=10)   
-    print('Accuracy with feature optimization: {}'.format(np.round(np.mean(cv['test_score']),3)))
-    """
 
 def Strawman_imputation(data):
     """
@@ -399,7 +476,7 @@ def KNN_imputation(data, imputer=None, k=3):
         The second output is the KNN Imputer that should be used to transform
         new data, prior to predictions. 
     """
-    
+
     if imputer is None:
         imputer = KNNImputer(n_neighbors=k)
         imputer.fit(data)
@@ -451,7 +528,7 @@ def MissForest_imputation(data):
         The second output is the Miss Forest Imputer that should be used to transform
         new data, prior to predictions. 
     """
-  
+
     if np.all(np.isfinite(data)):
         raise ValueError('No missing values in training dataset, do not apply MissForest imputation!')
     
@@ -460,3 +537,5 @@ def MissForest_imputation(data):
     imputed_data = imputer.transform(data)
 
     return imputer.transform(data) 
+
+
